#!/usr/bin/env python3
"""
Decompress a zlib compressed file
by Javantea
Feb 12, 2015

It turns out that rpyc files are compressed with zlib.
It outputs to stdout just like zcat.
TODO: Error checking
"""
from __future__ import print_function
import zlib
import sys

def usage():
	print("dezlib.py filename [...]")
	print("For example, git files are zlib compressed. So are renpy files.")
	print("./dezlib.py .git/objects/90/cfc130a8557392687f76965575c20ab3fd2e3a |less")
	print("./dezlib.py about.rpyc > about.rpyc.dec")
	return 1
#end def usage()

def main():
	if len(sys.argv) <= 1:
		return usage()
	#end if
	if '-h' in sys.argv or '--help' in sys.argv:
		return usage()
	#end if
	
	if sys.hexversion >= 0x3000000:
		outfunc = sys.stdout.buffer.write
	else:
		outfunc = sys.stdout.write
	#end if
	
	filenames = sys.argv[1:]
	for filename in filenames:
		a = open(filename,'rb')
		c = b'#'
		x = zlib.decompressobj()
		# One block at a time in case it's a gigabyte.
		while c != b'':
			c = a.read(4096)
			b = x.decompress(c)
			outfunc(b)
		#loop
		# Ensure that we got the last packet.
		b = x.flush()
		outfunc(b)
	#next filename
#end def main()

if __name__ == '__main__':
	main()
#end if
