#!/usr/bin/env python3
"""
Hammerwatch Unpacker
by Javantea
Nov 22, 2015

mkdirs from derenpy1.py
by Javantea

"""
from __future__ import print_function
import struct
import io
import os
import sys

def mkdirs(path):
	"""
	Ensure that a directory structure exists.
	FIXME: Deal with a file in the way.
	"""
	dirs = path.split(os.sep)
	dirx = '.' + os.sep
	for d in dirs:
		if not os.path.exists(dirx + d + os.sep):
			os.mkdir(dirx + d)
		#end if
		dirx += d + os.sep
	#next d
#end def mkdirs(path)

def hwra_unpack(a, verbose=True):
	"""
	Hammerwatch assets.bin unpacker
	"""
	if a[:4] != b'HWRA':
		print("Not a valid Hammerwatch assets.bin file.")
		return None
	#end if
	pos = 4
	files = struct.unpack('<I', a[pos:pos+4])
	print("files:", files)
	pos += 4
	while pos < len(a):
		filename_len = ord(a[pos:pos+1])
		pos += 1
		if pos >= len(a):
			print("End of file")
			break
		#end if
		filename = a[pos:pos+filename_len].decode('utf-8')
		pos += filename_len
		file_size = struct.unpack('<I', a[pos:pos+4])[0]
		pos += 4
		print(filename, 'len:', file_size)
		if '/' in filename:
			if '..' in filename or filename.startswith('/'):
				print("filename not sanitary.")
				pos += file_size
				continue
			#end if
			directory = 'data/' + os.sep.join(filename.split(os.sep)[:-1])
			if not os.path.exists(directory):
				mkdirs(directory)
			#end if
		#end if
		open('data/' + filename, 'wb').write(a[pos:pos+file_size])
		pos += file_size
	#loop
	return True
#end def hwra_unpack(a, verbose=True)

def main():
	filename = 'assets.bin'
	if len(sys.argv) > 1:
		filename = sys.argv[1]
	#end if
	# TODO: mmap or just read as you go. It's only 135 MB.
	a = open(filename, 'rb').read()
	if hwra_unpack(a) == True:
		exit(0)
	#end if
	exit(1)
#end def main()

if __name__ == '__main__':
	main()
#end if
