#!/usr/bin/env python3
"""
All Valid MP3s in One Binary
by Javantea
Sept 14, 2015

Number of valid MP3 headers.

# 207360 / 4294967296 = 0.005 %

The result from find_compress1.py: 207360 lines, which means one is missing =). 
The last one of course! Fixed.

The most interesting part about the resulting file is played back differently 
by 3 different mp3 decoders.
mplayer advertises 820 seconds of music but only plays 124 seconds.
mpg123 plays for a second.
audacity plays for 5 seconds.
firefox takes the mplayer approach.

Is this the start of an MP3 fuzzer? I suppose so.
valgrind --leak-check=full --log-file=valgrind_mplayer_mp3a.txt mplayer -really-quiet test/mp3_headers1.mp3
ERROR SUMMARY: 88 errors from 88 contexts (suppressed: 0 from 0)
"""
import find_compress1
import struct

HDR_SYNC = 0xffe00000
valids = []
# You could use 0x400000 or more, but due to header sync, it doesn't add 
# anything to valids. In fact the last valid is 0x1bebff. 0xfffbebff
for i in range(0x200000):
    s = struct.pack('>I', HDR_SYNC ^ i)
    if find_compress1.findMP3(s, 0): valids.append(i)
#next i

print('#', len(valids), '/', 1<<32, '=', '%3.6f %%' % (100 * len(valids)/(1<<32)))
for i in valids:
	print('%08x' % (HDR_SYNC ^ i))
#next i

open('mp3_headers1.bin', 'wb').write(b''.join([struct.pack('>I', HDR_SYNC ^ i) for i in valids]))
