#!/usr/bin/env python3
"""
Zlib Python
by Javantea
Sep 26, 2015 - Oct 6, 2015, original unsuccessful research about a decade ago.

A verbose and slow implementation of zlib in Python
This allows you to create new zlib streams from scratch.
"""
from __future__ import print_function
import copy
# we use zlib for adler32 only.
import zlib
import struct

class bitHandler:
	def __init__(self, f, pos):
		self.f = f
		# This might not be a good idea with mmap, but we'll see.
		self.flen = len(f)
		self.pos = pos
		self.bits = 0
		self.hold = 0
	#end def bitHandler(f, pos)
	
	def PULLBYTE(self):
		"""
		Grab a byte from the data and put it into the hold.
		"""
		# zlibPy.outputZlib(b'x\x9c5\x8b[\n\xc0 \x0c\x04\xffs+\x8f\xb2>JC\xad\x82hKo\xdf<\x10\x92\x9daC\x02\x81', 0)
		if self.pos >= self.flen: raise ValueError("unexpected end of data.")
		nextb = self.f[self.pos]
		# It's technically possible to get more than 32 bits but we don't 
		# care because our ints are big.
		self.hold += (nextb << self.bits)
		self.bits += 8
		self.pos += 1
	#end def PULLBYTE(self)
	
	def NEEDBITS(self, b):
		"""
		equivalent to while self.bits < b: self.PULLBYTE()
		"""
		#print("need bits", b, self.bits)
		if self.pos >= self.flen: raise ValueError("unexpected end of data.")
		while self.bits < b:
			nextb = self.f[self.pos]
			# It's technically possible to get more than 32 bits but we don't 
			# care because our ints are big.
			self.hold += (nextb << self.bits)
			self.bits += 8
			self.pos += 1
		#loop
	#end def NEEDBITS(b)
	
	def BITS(self, b):
		# No validation, you have to know what the world you're doing.
		return self.hold & ((1<<b)-1)
	#end def BITS(b)
	
	def DROPBITS(self, b):
		#print("drop bits", b, self.bits)
		# No validation, you have to know what the world you're doing.
		self.hold >>= b
		self.bits -= b
		if self.bits < 0: print("now you done it.")
	#end def DROPBITS(b)
	
	def have(self):
		"""
		Have is avail_in decremented by each pullbyte.
		"""
		# TODO: Validate
		return self.flen - self.pos
	#end def have()
	
	def left(self):
		"""
		avail_out, which we should always give them a lot of.
		"""
		return 0x1000000
	#end def left()
	
#end class bitHandler

# https://www.ietf.org/rfc/rfc1951.txt 3.2.5
# (extra_bits, min_value)
LENGTH_DECODER = [(0, 3), (0, 4), (0, 5), (0, 6), (0, 7), (0, 8), (0, 9), 
(0, 10), (1, 11), (1, 13), (1, 15), (1, 17), (2, 19), (2, 23), (2, 27), 
(2, 31), (3, 35), (3, 43), (3, 51), (3, 59), (4, 67), (4, 83), (4, 99), 
(4, 115), (5, 131), (5, 163), (5, 195), (5, 227), (0, 258), 
# Two extra for good measure...
(0, 0), (0, 0)]

DIST_DECODER = [(0, 1), (0, 2), (0, 3), (0, 4), (1, 5), (1, 7), (2, 9), 
(2, 13), (3, 17), (3, 25), (4, 33), (4, 49), (5, 65), (5, 97), (6, 129), 
(6, 193), (7, 257), (7, 385), (8, 513), (8, 769), (9, 1025), (9, 1537), 
(10, 2049), (10, 3073), (11, 4097), (11, 6145), (12, 8193), (12, 12289), 
(13, 16385), (13, 24577), 
# Two extra for good measure...
(0, 0), (0, 0)]

def reverseBits(v, l):
	return int(bin(v)[2:].rjust(l, '0')[::-1], 2)
#end def reverseBits(v, l)

ENOUGH_LENS = 852
ENOUGH_DISTS = 594
ENOUGH = (ENOUGH_LENS+ENOUGH_DISTS)
class inflate_state:
	def __init__(self, ncode, nlen, ndist, have):
		self.window = None
		self.ncode = ncode
		self.nlen = nlen
		self.ndist = ndist
		self.have = have
		self.inext = None
		self.lens = [0]*320
		self.work = [0]*288
		self.codes = [code() for i in range(ENOUGH)]
		
		self.lencode = None
		self.lenbits = 0
		self.mode = BAD
		self.back = 0
		self.dmax = 32768
	#end def inflate_state(ncode, nlen, ndist, have)
	
	def resetLens(self):
		# This could be cheaper, but we want it to be obvious...
		self.lens = [0]*320
	#end def resetLens()
#end class inflate_state

class code:
	def __init__(self, op=0, bits=0, val=0):
		self.op = op
		self.bits = bits
		self.val = val
	#end def code(op, bits, val)
	
	def __repr__(self):
		return 'code(op=%i, bits=%i, val=%i)' % (self.op, self.bits, self.val)
	#end def __repr__(self)
#end class code

class InflateTablesResp:
	def __init__(self, bits=0, table=None, table_used=0, work=None):
		self.bits = bits
		self.table = table
		self.table_used = table_used
		self.work = work
	#end def InflateTablesResp(bits=0, table=None, table_used=0, work=None)
#end class InflateTablesResp

CODES = 0
LENS = 1
DISTS = 2
MAXBITS = 13
def inflate_table(typet, lens, codes, table, table_offset, bits, work):
	here = code()
	next_sp = None
	offs = [0] * (MAXBITS + 1)
	count = [0] * (MAXBITS + 1)
	lbase = [ #/* Length codes 257..285 base */
		3, 4, 5, 6, 7, 8, 9, 10, 11, 13, 15, 17, 19, 23, 27, 31,
		35, 43, 51, 59, 67, 83, 99, 115, 131, 163, 195, 227, 258, 0, 0]
	lext = [ #* Length codes 257..285 extra */
		16, 16, 16, 16, 16, 16, 16, 16, 17, 17, 17, 17, 18, 18, 18, 18,
		19, 19, 19, 19, 20, 20, 20, 20, 21, 21, 21, 21, 16, 73, 195]
	dbase = [ #/* Distance codes 0..29 base */
		1, 2, 3, 4, 5, 7, 9, 13, 17, 25, 33, 49, 65, 97, 129, 193,
		257, 385, 513, 769, 1025, 1537, 2049, 3073, 4097, 6145,
		8193, 12289, 16385, 24577, 0, 0]
	dext = [ #/* Distance codes 0..29 extra */
		16, 16, 16, 16, 17, 17, 18, 18, 19, 19, 20, 20, 21, 21, 22, 22,
		23, 23, 24, 24, 25, 25, 26, 26, 27, 27,
		28, 28, 29, 29, 64, 64]
	# accumulate lengths for codes (assumes lens[] all in 0..MAXBITS)
	for sym in range(codes):
		#print('ls', lens[sym], count[lens[sym]])
		count[lens[sym]] += 1
	#next sym
	#print('count', count)
	# bound code lengths, force root to be within code lengths
	root = bits
	for maxb in range(MAXBITS, 0, -1):
		if count[maxb] != 0: break
	if root > maxb: root = maxb
	if maxb == 0:
		# Error condition.
		return None
	#end if
	minb = 1
	for minb in range(1, maxb):
		if count[minb] != 0: break
	if root < minb: root = minb
	#print('root', root, minb, maxb)
	# check for an over-subscribed or incomplete set of lengths
	left = 1
	for lenb in range(1, MAXBITS+1):
		left <<= 1
		left -= count[lenb]
		if left < 0:
			#print("e")
			return -1
	#next lenb
	if left > 0 and ((typet == CODES) or (maxb != 1)):
		#print("f", left, typet, maxb)
		return -1
	#end if
	
	# generate offsets into symbol table for each length for sorting
	offs[1] = 0
	for lenb in range(1, MAXBITS):
		offs[lenb + 1] = offs[lenb] + count[lenb]
	#next lenb
	
	# sort symbols by length, by symbol order within each length
	for sym in range(codes):
		if lens[sym] != 0:
			v = offs[lens[sym]]
			offs[lens[sym]] += 1
			work[v] = sym
		#end if
	#next sym
	
	# Create and fill in decoding tables. See zlib-1.2.5/inftrees.c for the 
	# full commentary.
	if typet == CODES:
		base = extra = work
		base_offset = 0
		extra_offset = 0
		end = 19
	elif typet == LENS:
		# This is impressive logic.
		base = lbase
		base_offset = 257
		extra = lext
		extra_offset = 256
		end = 256
	else: # DISTS
		base_offset = 0
		extra_offset = 0
		base = dbase
		extra = dext
		end = -1
	#end if
	
	# initialize state for loop
	huff = 0
	sym = 0
	lenb = minb
	next_sp = 0
	curr = root
	drop = 0
	low = 0x7fffffff #(unsigned)(-1)
	used = 1 << root
	mask = used - 1
	
	# check available table space
	if typet == LENS and used >= ENOUGH_LENS:
		#print("d")
		return 1
	if typet == DISTS and used >= ENOUGH_DISTS:
		#print("c")
		return 1
	
	# process all codes and make table entries
	while True:
		# create table entry
		here.bits = (lenb - drop)
		if work[sym] < end:
			#print('t1', sym, work[sym], end)
			here.op = 0
			here.val = work[sym]
		elif work[sym] > end:
			#print('t2', sym, work[sym], end)
			here.op = extra[work[sym]-extra_offset]
			here.val = base[work[sym]-base_offset]
		else:
			#print('t3', sym, work[sym], end)
			here.op = 32 + 64 # end of block
			here.val = 0
		#end if

		# replicate for those indices with low len bits equal to huff
		incr = 1 << (lenb - drop)
		fill = 1 << curr
		minx = fill # save offset to next table
		FirstRound = True
		while FirstRound or fill != 0:
			fill -= incr
			# next[(huff >> drop) + fill] = here;
			#print('fill', table_offset + next_sp + (huff >> drop) + fill, here.val, table_offset)
			table[table_offset + next_sp + (huff >> drop) + fill] = copy.copy(here)
			FirstRound = False
		#while (fill != 0)

		# backwards increment the len-bit code huff
		incr = 1 << (lenb - 1)
		while (huff & incr):
			incr >>= 1
		#loop
		if incr != 0:
			huff &= incr - 1
			huff += incr
		else:
			huff = 0
		#end if

		# go to next symbol, update count, len
		sym += 1
		count[lenb] -= 1
		if count[lenb] == 0:
			# The only non-error way to get out of this endless loop!
			if lenb == maxb: break
			lenb = lens[work[sym]]
		#end if

		# create new sub-table if needed
		if lenb > root and (huff & mask) != low:
			# if first time, transition to sub-tables
			if drop == 0:
				drop = root
			#end if

			# increment past last table
			next_sp += minb            # here min is 1 << curr

			# determine length of next table
			curr = lenb - drop
			left = 1 << curr
			while (curr + drop < maxb):
				left -= count[curr + drop]
				if left <= 0: break
				curr += 1
				left <<= 1
			# loop

			# check for enough space
			used += 1 << curr
			if typet == LENS and used >= ENOUGH_LENS:
				#print("a")
				return 1
			if typet == DISTS and used >= ENOUGH_DISTS:
				#print("b")
				return 1

			# point entry in root table to sub-table
			low = huff & mask
			#print('low', table_offset + low, next_sp)
			table[table_offset + low] = code(curr, root, next_sp)
		#end if
	#loop
	
	# Fill in rest of table for incomplete codes. See zlib-1.2.5/inftrees.c for 
	# the full commentary.
	here.op = 64
	here.bits = lenb - drop
	here.val = 0
	while huff != 0:
		# when done with sub-table, drop back to root table
		if (drop != 0 and (huff & mask) != low):
			drop = 0
			lenb = root
			next_sp = 0
			here.bits = lenb
		#end if

		# put invalid code marker in table
		table[next_sp + (huff >> drop)] = here

		# backwards increment the len-bit code huff
		incr = 1 << (lenb - 1)
		while (huff & incr):
			incr >>= 1
		#loop
		if (incr != 0):
			huff &= incr - 1
			huff += incr
		else:
			huff = 0
		#end if
	#loop

	# set return parameters
	# This is a very odd use of pointers...
	#*table += used
	return InflateTablesResp(root, table, table_offset + used, work)
#end def inflate_table(typet, lens, codes, table, table_offset, bits, work)

def inflate_fast(bh, out):
	"""
	FIXME: You wanted it to go fast.
	"""
	return 0
#end def inflate_fast(bh, out)

# enum inflate_mode, see inflate.h
HEAD=0     # i: waiting for magic header */
FLAGS=1    # i: waiting for method and flags (gzip) */
TIME=2     # i: waiting for modification time (gzip) */
OS=3       # i: waiting for extra flags and operating system (gzip) */
EXLEN=4    # i: waiting for extra length (gzip) */
EXTRA=5    # i: waiting for extra bytes (gzip) */
NAME=6     # i: waiting for end of file name (gzip) */
COMMENT=7  # i: waiting for end of comment (gzip) */
HCRC=8     # i: waiting for header crc (gzip) */
DICTID=9   # i: waiting for dictionary check value */
DICT=10    # waiting for inflateSetDictionary() call */
TYPE=11      # i: waiting for type bits  including last-flag bit */
TYPEDO=12    # i: same  but skip check to exit inflate on new block */
STORED=13    # i: waiting for stored size (length and complement) */
COPY_=14     # i/o: same as COPY below  but only first time in */
COPY=15      # i/o: waiting for input or output to copy stored block */
TABLE=16     # i: waiting for dynamic block table lengths */
LENLENS=17   # i: waiting for code length code lengths */
CODELENS=18  # i: waiting for length/lit and distance code lengths */
LEN_=19        # i: same as LEN below  but only first time in */
LEN=20         # i: waiting for length/lit/eob code */
LENEXT=21      # i: waiting for length extra bits */
DIST=22        # i: waiting for distance code */
DISTEXT=23     # i: waiting for distance extra bits */
MATCH=24       # o: waiting for output space to copy string */
LIT=25         # o: waiting for output space to write literal */
CHECK=26   # i: waiting for 32-bit check value */
LENGTH=27  # i: waiting for 32-bit length (gzip) */
DONE=28    # finished check  done -- remain here until reset */
BAD=29     # got a data error -- remain here until reset */
MEM=30     # got an inflate() memory error -- remain here until reset */
SYNC=40    # looking for synchronization bytes to restart inflate() */

def Zlib_InflateSlow(state, bh, window):
	"""
	A nearly as fast as inflate_fast implementation of zlib dynamic codes.
	Okay, it's very slow. What can I say?
	TODO: Verify extra bits which aren't used very often in legitimate zlib.
	"""
	# LEN
	# inflate_fast is a bit bigger, so I'm going to shelve it.
	#if bh.have() >= 6 and bh.left() >= 258:
	#	#bh.RESTORE()
	#	# FIXME: Dunno what to put here.
	#	out = 80000
	#	inflate_fast(bh, out)
	#	#bh.LOAD()
	#	#if state.mode == TYPE:
	#	#   state.back = -1
	#	break
	#end if
	# TODO: verify there's enough data in our bh to go on.
	# zlibPy.outputZlib(b'x\x9c5\x8b[\n\xc0 \x0c\x04\xffs+\x8f\xb2>JC\xad\x82hKo\xdf<\x10\x92\x9daC\x02\x81', 0)
	state.back = 0
	while True:
		here = state.codes[state.lencode + bh.BITS(state.lenbits)]
		if here.bits <= bh.bits: break
		bh.PULLBYTE()
	#loop
	#print('endme:', bh.BITS(state.lenbits))
	#print('h', here.op, here.bits, here.val, state.lenbits)
	if here.op and ((here.op & 0xf0) == 0):
		last = here
		while True:
			here = state.codes[state.lencode + last.val + 
				(bh.BITS(last.bits + last.op) >> last.bits)]
			if here.bits <= bh.bits: break
			bh.PULLBYTE()
		#loop
		bh.DROPBITS(last.bits)
		state.back += last.bits
	#end if
	bh.DROPBITS(here.bits)
	state.back += here.bits
	state.length = here.val
	if here.op == 0:
		v = bytes([here.val])
		print("inflate:         literal %r" % v)
		state.mode = LIT
		# ...
		state.mode = LEN
		return v
	#end if
	if here.op & 32:
		print("inflate:         end of block")
		state.mode = TYPE
		return b''
	#end if
	if here.op & 64:
		# Error state
		# FIXME: zlibPy.outputZlib(b'x\x9c\xed\xc11\t\x00\x00\x0c\x030c\xa30\xff\xa6j\xa3G\x92{\x00\x00`B\nwr\xf35', 0)
		print('h', here.op, here.bits, here.val, state.lenbits)
		print("invalid literal/length code")
		state.mode = BAD
		return b''
	#end if
	state.extra = here.op & 15
	state.mode = LENEXT
	# LENEXT
	if state.extra:
		bh.NEEDBITS(state.extra)
		state.length += bh.BITS(state.extra)
		bh.DROPBITS(state.extra)
		state.back += state.extra
	#end if
	state.was = state.length
	state.mode = DIST
	# DIST
	#print('codes3:', len(state.codes), state.codes)
	#print('bhb', bh.bits)
	bh.PULLBYTE()
	while True:
		# This seems broken, but I'm going to allow it until...
		#print('dist', state.distcode, bh.BITS(state.distbits), state.distbits)
		here = state.codes[state.distcode + bh.BITS(state.distbits)]
		if here.bits <= bh.bits: break
		bh.PULLBYTE()
	#loop
	#print('hd', here.op, here.bits, here.val)
	if (here.op & 0xf0) == 0:
		last = here
		while True:
			here = state.codes[state.distcode + last.val +
					(bh.BITS(last.bits + last.op) >> last.bits)]
			if (last.bits + here.bits) <= bh.bits: break
			bh.PULLBYTE()
		#loop
		bh.DROPBITS(last.bits)
		state.back += last.bits
	#end if
	bh.DROPBITS(here.bits)
	state.back += here.bits
	if here.op & 64:
		print("invalid distance code")
		state.mode = BAD
		return b''
	#end if
	state.offset = here.val
	state.extra = here.op & 15
	state.mode = DISTEXT
	# DISTEXT
	if state.extra:
		print("could you spare some extra bits?", state.extra)
		bh.NEEDBITS(state.extra)
		state.offset += bh.BITS(state.extra)
		bh.DROPBITS(state.extra)
		state.back += state.extra
	#end if
	if state.offset > state.dmax:
		print("invalid distance too far back")
		state.mode = BAD
		return b''
	#end if
	print("inflate:         distance %u" % state.offset, end=' ')
	state.mode = MATCH
	# MATCH
	#print('window:', window)
	state.mode = LEN
	# I'm smart enough to do this myself =)
	output = window[-state.offset:]
	# This is a nifty little error.
	if len(output) == 0: return b''
	while len(output) < state.length:
		output += window[-state.offset:]
	#loop
	print(repr(output[:state.length]))
	return output[:state.length]
#end def Zlib_InflateSlow(state, bh, window)

def outputZlib(f, fpos):
	"""
	Given N bytes of data that passes findZlib, find the end and parts of a 
	Zlib.
	
	"""
	data = f[fpos:fpos+10]
	if data == None or len(data) < 2: return False
	# Pretty much all zlib is 'x', anomalies include 'h' and so forth.
	# 
	if (data[0]) & 0xf != 8: return False
	CMF = (data[0])
	FLG = (data[1])
	# Easy FCHECK
	mod31_fcheck = ((CMF<<8) + FLG) % 31
	if mod31_fcheck != 0: return False

	state = inflate_state(0, 0, 0, 0)
	CM = (CMF & 0xf)
	# Already checked CM above
	CINFO = (CMF & 0xf0) >> 4
	# This is a limitation of zlib, https://www.ietf.org/rfc/rfc1950.txt
	if CINFO > 7:
		return False
	#end if
	# Almost invariably 32768.
	window_size = 1 << (CINFO+8)
	state.dmax = window_size

	pos = 2
	fdict = (FLG & 0x20) >> 5
	dictid = None
	if fdict == 1:
		# TODO: Use DICTID maybe?
		dictid = data[2:6]
		pos += 4
	#end if
	# Worthless data. 0 = fastest, 1 = fast, 2 = default, 3 = slowest
	flevel = (FLG & 0xc0) >> 6
	
	flen = len(f)
	if len(data) <= pos: return data[:pos]

	# Now pos is not talking about data, it's talking about f.
	pos += fpos
	bh = bitHandler(f, pos)
	#decompressed = b''
	window = b''
	adler_value = 1
	bfinal = 0
	state.mode = TYPEDO
	while bfinal == 0:
		# zlibPy.outputZlib(b'8O2', 0)
		if bh.pos+1 >= flen: return f[fpos:pos]
		# DEFLATE
		# BTYPE reserved (error) https://www.ietf.org/rfc/rfc1951.txt pp. 9
		bh.NEEDBITS(3)
		# bfinal signals the last block. 
		bfinal = bh.BITS(1)
		bh.DROPBITS(1)
		btype = bh.BITS(2)
		bh.DROPBITS(2)
		# This is an error condition, we should return False, but we want to give them the stuff that succeeded.
		if btype == 0x3:
			print("Bug or bad data", btype)
			return f[fpos:pos]
		#end if
		if btype == 0x0:
			# stored block
			# find_compress1.outputZlib(zlib.compress(zlib.compress(open('/usr/share/dict/words','rb').read(100))), 0)
			# Who decided that it should use little endian?!
			# find_compress1.outputZlib(b'80aaaaa', 0)
			if bh.pos+5 > flen:
				print("Unexpected end of data")
				return f[fpos:pos]
			#end if
			block_len = struct.unpack('<H', f[pos+1:pos+3])[0]
			block_nlen = struct.unpack('<H', f[pos+3:pos+5])[0]
			print('stored block:', block_len, block_nlen)
			if (block_len + block_nlen) != 0xffff:
				print("Bad stored block")
				return f[fpos:pos]
			#end if
			v = f[pos+5:pos+5+block_len]
			pos += 5 + block_len
			bh.pos = pos
			adler_value = zlib.adler32(v, adler_value)
		elif btype == 1:
			# fixed codes block
			# find_compress1.outputZlib(zlib.compress(b'asdf1234'), 0)
			# Fixed codes are so wasteful I can only imagine them being used with lame blocks.
			#   0-143 8  00110000  10111111
			# 144-255 9 110010000 111111111
			# 256-279 7   0000000   0010111
			# 280-287 8  11000000  11000111
			# See zlib1.txt for more detail on how this occurred.
			print("btype 1 fixed codes")
			# Grab the 5 msb of the value
			"""
			00110
			10111
			11001
			11111
			00000
			00101
			11000
			11000
			"""
			while True:
				# Might cause a problem... but due to adler32 we may have padding.
				# zlibPy.outputZlib(b'8O2', 0)
				# zlibPy.outputZlib(b'8O2aaaa', 0)
				if bh.pos+2 >= flen: return f[fpos:pos]
				bh.NEEDBITS(9)
				test1 = bh.BITS(5)
				test1b = bin(test1)[2:].rjust(5, '0')[::-1]
				if test1 == 3: #test1b == '11000':
					# 280-287 8  11000000  11000111
					#print(bin(bh.BITS(8))[2:].rjust(8, '0')[::-1])
					code = reverseBits(bh.BITS(8), 8)
					decode = (code - 0b11000000) + 280
					bh.DROPBITS(8)
				elif test1 & 0x3 == 3:
					# 144-255 9 110010000 111111111
					#print(bin(bh.BITS(9))[2:].rjust(9, '0')[::-1])
					code = reverseBits(bh.BITS(9), 9)
					decode = (code - 0b110010000) + 144
					bh.DROPBITS(9)
				# TODO: Reduce this.
				elif test1 == 0 or test1 == 0b10000 or test1 == 0b01000 or test1 == 0b11000 or test1 == 0b00100 or test1 == 0b10100:
				#elif test1b == '00000' or test1b == '00001' or test1b == '00010' or test1b == '00011' or test1b == '00100' or test1b == '00101':
					# 256-279 7   0000000   0010111
					#print(bin(bh.BITS(7))[2:].rjust(7, '0')[::-1])
					code = reverseBits(bh.BITS(7), 7)
					bh.DROPBITS(7)
					# End of data
					if code == 0: break
					decode = (code) + 256
				else:
					#   0-143 8  00110000  10111111
					#print(bin(bh.BITS(8))[2:].rjust(8, '0')[::-1])
					code = reverseBits(bh.BITS(8), 8)
					decode = (code - 0b00110000)
					bh.DROPBITS(8)
				#end if
				#decompressed += bytes([decode])
				if decode <= 255:
					v = bytes([decode])
					window += v
				else:
					# It's a length.
					v = b''
					dec = LENGTH_DECODER[decode - 257]
					if dec[0] > 0:
						# TODO: Test this out, see what happens.
						bh.NEEDBITS(dec[0])
						extra_bits = bh.BITS(dec[0])
						bh.DROPBITS(dec[0])
						lengthValue = dec[1] + extra_bits
					else:
						lengthValue = dec[1]
					#end if
					#print('lv', lengthValue)
					# The next will be a dist, so grab it.
					# Turns out 6 fits b'Blah blahtxyz'
					bh.NEEDBITS(5)
					distCode = reverseBits(bh.BITS(5), 5)
					bh.DROPBITS(5)
					"""
					Distance codes 0-31 are represented by (fixed-length) 5-bit
					codes, with possible additional bits as shown in the table
					shown in Paragraph 3.2.5, above.  Note that distance codes 30-
					31 will never actually occur in the compressed data.
					"""
					distDec = DIST_DECODER[distCode]
					if distDec[0] > 0:
						bh.NEEDBITS(distDec[0])
						extra_dist_bits = bh.BITS(distDec[0])
						#print('edb', extra_dist_bits)
						bh.DROPBITS(distDec[0])
						distValue = distDec[1] + extra_dist_bits
					else:
						distValue = distDec[1]
					#end if
					#print('ld', lengthValue, distCode, distValue)
					pos = len(window) - distValue
					for i in range(lengthValue):
						new_byte = window[pos+i:pos+i+1]
						window += new_byte
						v += new_byte
					#next i
				#end if
				print(code, decode, hex(decode), v)
				# byte by byte adler32.
				adler_value = zlib.adler32(v, adler_value)
				# expensive with large sets of data. but not expensive enough 
				# for me to care.
				if len(window) > window_size:
					window = window[-window_size:]
				#end if
			#loop
			#print('now at', bh.pos)
			pos = bh.pos
		elif btype == 2:
			# dynamic codes
			# find_compress1.outputZlib(zlib.compress(open('/usr/share/dict/words','rb').read(150)), 0)
			# Very commonly used in larger compressed streams.
			# zlibPy.outputZlib(b'8O4aaaa', 0)
			if bh.pos+2 >= flen: return f[fpos:pos]
			bh.NEEDBITS(14)
			state.nlen = bh.BITS(5) + 257 # LIT
			bh.DROPBITS(5)
			state.ndist = bh.BITS(5) + 1 # DIST
			bh.DROPBITS(5)
			state.ncode = bh.BITS(4) + 4 # CLEN
			bh.DROPBITS(4)
			print('dynamic codes:', state.nlen, state.ndist, state.ncode)
			"""
			Alphabet for code lengths:
			0 - 15: code lengths 0 - 15
			    16: copy the previous code length 3 - 6 times.
			        read 2 bits to find out...
			    17: Repeat a code length of 0 for 3 - 10 times.
			        read 3 bits to find out...
			    18: Repeat a code length of 0 for 11 - 138 times.
			        read 7 bits to find out...
			        
			static const unsigned short order[19] = /* permutation of code lengths */
			"""
			state.have = 0
			order = [16, 17, 18, 0, 8, 7, 9, 6, 10, 5, 11, 4, 12, 3, 13, 2, 14, 1, 15]
			state.resetLens()
			if bh.pos+((state.ncode*3)//8) >= flen: return f[fpos:pos]
			for i in range(state.ncode):
				bh.NEEDBITS(3)
				state.lens[order[i]] = bh.BITS(3)
				bh.DROPBITS(3)
				#print('l', state.lens[order[i]])
				state.have += 1
			#next i
			#print('code lengths:', state.lens, state.have)
			state.have = 19
			state.nexti = 0
			state.lencode = state.nexti
			state.lenbits = 7
			ret = inflate_table(CODES, state.lens, 19, state.codes, 
				state.nexti, state.lenbits, state.work)
			if not isinstance(ret, InflateTablesResp):
				# Error in inflate_table
				print(" invalid code lengths set")
				return f[fpos:pos]
			#end if
			print("inflate:       code lengths ok")
			state.lenbits = ret.bits
			# This gets written by passing in state.codes.
			#state.codes = state.codes[:state.nexti] + ret.table[state.nexti:]
			state.nexti += ret.table_used
			# This gets written by passing in state.work.
			#state.work = ret.work
			#print('codes1:', len(state.codes), state.codes[:ret.table_used], set(state.codes[ret.table_used:]), state.nexti)
			state.have = 0
			#state.mode = CODELENS
			# CODELENS
			clen = 0
			while state.have < (state.nlen + state.ndist):
				while True:
					# This is a mind-bender!
					here = state.codes[state.lencode + bh.BITS(state.lenbits)]
					#print('mindbend', here.bits, bh.bits, state.lenbits, bh.BITS(state.lenbits))
					if here.bits <= bh.bits: break
					bh.PULLBYTE()
				#loop
				if here.val < 16:
					bh.NEEDBITS(here.bits)
					bh.DROPBITS(here.bits)
					state.lens[state.have] = here.val
					#print('writing', state.have, here.val)
					state.have += 1
				else:
					if here.val == 16:
						bh.NEEDBITS(here.bits + 2)
						bh.DROPBITS(here.bits)
						if state.have == 0:
							#"invalid bit length repeat";
							return f[fpos:pos]
						#end if
						clen = state.lens[state.have - 1]
						copy = 3 + bh.BITS(2)
						bh.DROPBITS(2)
					elif here.val == 17:
						bh.NEEDBITS(here.bits + 3)
						bh.DROPBITS(here.bits)
						clen = 0
						copy = 3 + bh.BITS(3)
						bh.DROPBITS(3)
					else:
						bh.NEEDBITS(here.bits + 7)
						bh.DROPBITS(here.bits)
						clen = 0
						copy = 11 + bh.BITS(7)
						bh.DROPBITS(7)
					#end if
					if state.have + copy > state.nlen + state.ndist:
						#strm->msg = (char *)"invalid bit length repeat";
						return f[fpos:pos]
					#end if
					while copy:
						copy -= 1
						state.lens[state.have] = clen
						#print('copying', state.have, clen)
						state.have += 1
					#end if
					# Not sure if this is necessary...
					copy -= 1
				#end if
			#loop
			if state.lens[256] == 0:
				# invalid code -- missing end-of-block
				print("invalid code -- missing end-of-block")
				return f[fpos:pos]
			#end if
			#print('end code', state.lens[256])
			
			# build code tables -- ...
			state.nexti = 0
			state.lencode = state.nexti
			state.lenbits = 9
			ret = inflate_table(LENS, state.lens, state.nlen, state.codes, 
				state.nexti, state.lenbits, state.work)
			if not isinstance(ret, InflateTablesResp):
				# Error in inflate_table
				# invalid literal/lengths set
				return f[fpos:pos]
			#end if
			state.lenbits = ret.bits
			# This gets written by passing in state.codes.
			#state.codes = state.codes[:state.nexti] + ret.table[state.nexti:]
			state.nexti += ret.table_used
			# This gets written by passing in state.work.
			#state.work = ret.work

			state.distcode = state.nexti
			state.distbits = 6
			ret = inflate_table(DISTS, state.lens[state.nlen:], state.ndist,
				state.codes, state.nexti, state.distbits, state.work)
			if not isinstance(ret, InflateTablesResp):
				# Error in inflate_table
				# invalid distances set
				return f[fpos:pos]
			#end if
			state.distbits = ret.bits
			#print('bug', state.nexti, state.codes[:4])
			# This gets written by passing in state.codes.
			#state.codes = state.codes[:state.nexti] + ret.table[state.nexti:]
			state.nexti += ret.table_used
			# This gets written by passing in state.work.
			#state.work = ret.work
			print("inflate:       codes ok")
			#print('codes2:', len(state.codes), state.codes[:ret.table_used], set(state.codes[ret.table_used:]), state.nexti)
			# state->mode = LEN_;
			# LEN_
			# Find the end code for our zlib tattoo research.
			for i, c in enumerate(state.codes):
				if c.op & 32:
					print("End code:", i, c)
					break
				#end if
			#next i, c
			state.mode = LEN
			# LEN and so on is taken care of by this function.
			# since state and bh are passed byref, we're okay when we get them back.
			i = 0
			while state.mode == LEN:
				v = Zlib_InflateSlow(state, bh, window)
				if v:
					adler_value = zlib.adler32(v, adler_value)
					window += v
				#end if
				i += 1
				#if i > 20: break
			#loop
			print("End state:", state.mode)
			if state.mode == BAD:
				# Error.
				return f[fpos:pos]
			#end if
			pos = bh.pos
		#end if
	#loop
	pos -= (bh.bits >> 3)
	# Now for the adler
	adler_read = None
	if len(f) >= pos+4:
		adler_read = struct.unpack('>I', f[pos:pos+4])[0]
		pos += 4
	#end if
	# TODO: on large data, you don't want to do this all at once, right? also...
	if adler_value != adler_read:
		print("Warning: didn't decompress properly according to adler32.", adler_value)
	#end if
	print(window)
	return f[fpos:pos]
#end def outputZlib(f, fpos)

